from sqlalchemy import (
    Column,
    Enum,
    ForeignKeyConstraint,
    Integer,
    MetaData,
    String,
    Table,
)

from ...dto import Messengers

metadata = MetaData()


dialogs = Table(
    "dialogs",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("messenger", Enum(Messengers), primary_key=True),
    Column("username", String),
)


dialog_steps = Table(
    "dialog_steps",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("dialog_id", Integer),
    Column("messenger", Enum(Messengers)),
    Column("question", String),
    Column("answer", String),
    ForeignKeyConstraint(
        ["dialog_id", "messenger"], ["dialogs.id", "dialogs.messenger"]
    ),
)



from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import create_async_engine

from ..archetype import ArchetypeStorage
from . import tables


class PostgreStorage(ArchetypeStorage):
    def __init__(self, uri):
        self._engine = create_async_engine(uri)

    async def save_question_and_answer(self, dialog, question):
        async with self._engine.begin() as conn:
            await conn.execute(
                tables.dialog_steps.insert().values(
                    {
                        "dialog_id": dialog.identifier,
                        "messenger": dialog.service.type,
                        "question": question.topic,
                        "answer": dialog.answer.text,
                    }
                )
            )

    async def save_dialog(self, dialog):
        async with self._engine.begin() as conn:
            try:
                await conn.execute(
                    tables.dialogs.insert().values(
                        {
                            "id": dialog.identifier,
                            "messenger": dialog.service.type,
                            "username": dialog.username,
                        }
                    )
                )
            except IntegrityError:
                pass



from abc import ABCMeta, abstractmethod
from asyncio import Event, wait_for
from copy import copy

from .. import const
from ..dto import Answer, Message
from ..exceptions import QuestionWrongAnswer
from ..question import Question


class ArchetypeService(metaclass=ABCMeta):
    def __init__(self, quiz, storage, settings, *args, **kwargs):
        self.stop_event = Event()
        self.dialogs = {}
        self.quiz = quiz
        self.storage = storage
        self.settings = settings

    @property
    @abstractmethod
    def type(self):
        pass

    def stop(self):
        self.stop_event.set()

    @abstractmethod
    def run_forever(self):
        pass

    @abstractmethod
    async def send_message(self, user_id, *args, **kwargs):
        pass


class ArchetypeDialog(metaclass=ABCMeta):
    def __init__(
        self,
        service: ArchetypeService,
        identifier: int,
        username: str,
        answer_timeout: int = const.ANSWER_TIMEOUT,
    ):
        self.service = service
        self.identifier = identifier
        self.username = username
        self.last_question_id = 0
        self.answer = Answer()
        self.answer_timeout = answer_timeout

    @abstractmethod
    def prepare_question(self, question: Question) -> dict:
        pass

    @abstractmethod
    def prepare_answer(self, question: Question, answer: Answer) -> Answer:
        pass

    async def ask(self, question: Question) -> Answer:
        self.answer.clear()

        message_data = self.prepare_question(question)

        self.last_question_id = await self.tell(**message_data)

        while 1:
            try:
                await wait_for(self.answer.wait(), self.answer_timeout)

                self.answer = self.prepare_answer(question, self.answer)

                question.validate_answer(self.answer)

                await self.service.storage.save_question_and_answer(self, question)

                return copy(self.answer)
            except QuestionWrongAnswer:
                await self.tell(const.WRONG_ANSWER_FORMAT)
                self.answer.clear()

    async def tell(self, *args, **kwargs) -> int:
        return await self.service.send_message(self.identifier, *args, **kwargs)

    async def handle_message(self, message: Message):
        if message.id < self.last_question_id:
            return

        self.answer.set(message.text)

    async def start(self):
        await self.service.storage.save_dialog(self)

    async def close(self):
        pass



import logging

from telethon import Button, TelegramClient, events

from .. import const
from ..dto import Message, Messengers
from .archetype import ArchetypeDialog, ArchetypeService


class TelegramDialog(ArchetypeDialog):
    def prepare_question(self, question) -> dict:
        message = question.topic

        buttons = []

        if question.options:
            for index, text in enumerate(question.options, 1):
                if question.inline:
                    buttons.append(Button.inline(text, index))
                else:
                    buttons.append(Button.text(text, single_use=True))

            if len(buttons) > question.column_count:
                rows_buttons = list(zip(*(question.column_count * [iter(buttons)])))
                if len(buttons) % question.column_count:
                    rows_buttons.append(
                        buttons[-(len(buttons) % question.column_count) :]
                    )
                buttons = rows_buttons

            return {"message": message, "buttons": buttons}

        return {"message": message}

    def prepare_answer(self, question, answer):
        if question.options and question.inline:
            try:
                index = int(answer.text) - 1
                answer.text = question.options[index]
            except ValueError:
                pass

        return answer


class TelegramService(ArchetypeService):
    type = Messengers.telegram

    def __init__(self, quiz, storage, settings, *args, **kwargs):
        super().__init__(quiz, storage, settings)
        self._bot = TelegramClient(
            "testbot", settings["api_id"], settings["api_hash"], proxy=None
        ).start(bot_token=settings["token"])

    async def handle_click_button(self, event):
        if event.chat_id not in self.dialogs:
            await self._bot.send_message(event.chat, const.FOREWORD)
            return

        dialog = self.dialogs[event.chat_id]

        message = Message(event.message_id, event.query.data.decode())
        await dialog.handle_message(message)

    async def handle_new_message(self, event):
        if event.chat_id not in self.dialogs:
            await self._bot.send_message(event.chat, const.FOREWORD)
            return

        dialog = self.dialogs[event.chat_id]

        Message(event.message.id, event.message.text)
        await dialog.handle_message(event.message)

    async def handle_start(self, event):
        """Start super!"""

        try:
            if event.chat_id in self.dialogs:
                await self.close_dialog(event.chat)

            dialog = await self.create_dialog(event.chat)
            await self.quiz(dialog)
        finally:
            await self.close_dialog(event.chat)
            raise events.StopPropagation

    async def handle_cancel(self, event):
        try:
            if event.chat_id not in self.dialogs:
                await self._bot.send_message(event.chat, const.SESSION_DOESNT_EXIST)
                return

            await self.close_dialog(event.chat)
            await self._bot.send_message(event.chat, const.CANCEL)
        finally:
            raise events.StopPropagation

    async def close_dialog(self, user):
        dialog = self.dialogs.pop(user.id, None)
        if dialog:
            await dialog.close()
            logging.info("Dialog {} was closed".format(user.id))
        else:
            logging.info("Dialog {} doesn't found".format(user.id))

    async def create_dialog(self, user):
        dialog = TelegramDialog(
            self, user.id, user.username, **self.settings.get("dialog", {})
        )

        await dialog.start()

        self.dialogs[user.id] = dialog

        logging.info("New dialog was created with id: {}".format(user.id))

        return dialog

    async def send_message(self, user_id, *args, **kwargs):
        message = await self._bot.send_message(user_id, *args, **kwargs)
        return message.id

    def run_forever(self):
        self._bot.add_event_handler(
            self.handle_start, events.NewMessage(pattern="/start")
        )
        self._bot.add_event_handler(
            self.handle_cancel, events.NewMessage(pattern="/cancel")
        )
        self._bot.add_event_handler(self.handle_new_message, events.NewMessage)
        self._bot.add_event_handler(self.handle_click_button, events.CallbackQuery)
        self._bot.run_until_disconnected()



import enum
from asyncio import Event
from dataclasses import dataclass


@dataclass
class Message:
    id: int
    text: str


class Answer(Event):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text = None

    def clear(self):
        self.text = None
        return super().clear()

    def set(self, text):
        self.text = text
        return super().set()


class Messengers(enum.Enum):
    telegram = 1
    viber = 2
    whatapp = 3



from abc import ABCMeta, abstractmethod


class ArchetypeStorage(metaclass=ABCMeta):
    @abstractmethod
    async def save_question_and_answer(self, dialog, question):
        pass

    @abstractmethod
    async def save_dialog(self, dialog):
        pass



import typing

from .dto import Answer
from .exceptions import (
    QuestionChoicesWrongType,
    QuestionParameterWrongType,
    QuestionWrongAnswer,
)


class ANY:
    pass


class Question:
    def __init__(
        self,
        topic: str,
        choices: typing.Union[
            typing.Dict[typing.Any, str], typing.List[str], ANY
        ] = ANY,
        strict_choose: bool = True,
        column_count: int = 2,
        inline: bool = True,
    ):
        if not isinstance(topic, str):
            raise QuestionParameterWrongType(
                "Field `topic`: is expected type `str`, received: {}".format(
                    type(topic)
                )
            )

        if not isinstance(strict_choose, bool):
            raise QuestionParameterWrongType(
                "Field `strict_choose` is expected type `bool`, received: {}".format(
                    type(strict_choose)
                )
            )

        if not isinstance(column_count, int):
            raise QuestionParameterWrongType(
                "Field `column_count`: is expected type `int`, received: {}".format(
                    type(column_count)
                )
            )

        #  Check that choices is ANY or dict/list of string, otherwise raise Questionexception
        if not self._validate_choices(choices):
            raise QuestionChoicesWrongType(
                "Non empty dict or list of string or ANY is expected, received: {}".format(
                    choices
                )
            )

        self.topic = topic
        self.choices = choices
        self.strict_choose = strict_choose
        self.column_count = column_count
        self.inline = inline

    @property
    def options(self):
        if isinstance(self.choices, dict):
            return list(self.choices.values())
        elif isinstance(self.choices, list):
            return self.choices

    def _validate_choices(self, choices) -> bool:
        if choices == ANY:
            return True
        if choices and isinstance(choices, dict):
            iterator = (value for _, value in choices.items())
        elif choices and isinstance(choices, list):
            iterator = choices
        else:
            return False

        for value in iterator:
            if not isinstance(value, str):
                return False

        return True

    def validate_answer(self, answer: Answer):
        if (
            self.strict_choose
            and self.choices != ANY
            and answer.text not in self.options
        ):
            raise QuestionWrongAnswer(
                "Question: {}\nchoose invalid answer: {}".format(
                    self.topic, answer.text
                )
            )

        return True



"""Initial metadata

Revision ID: cfae1eb34e12
Revises:
Create Date: 2020-11-29 01:42:48.998482

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "cfae1eb34e12"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "dialogs",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column(
            "messenger",
            sa.Enum("telegram", "viber", "whatapp", name="messengers"),
            nullable=False,
        ),
        sa.Column("username", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id", "messenger"),
    )
    op.create_table(
        "dialog_steps",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("dialog_id", sa.Integer(), nullable=True),
        sa.Column(
            "messenger",
            sa.Enum("telegram", "viber", "whatapp", name="messengers"),
            nullable=True,
        ),
        sa.Column("question", sa.String(), nullable=True),
        sa.Column("answer", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ["dialog_id", "messenger"],
            ["dialogs.id", "dialogs.messenger"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("dialog_steps")
    op.drop_table("dialogs")
    # ### end Alembic commands ###



from .archetype import ArchetypeDialog, ArchetypeService


class ViberDialog(ArchetypeDialog):
    def prepare_message(self, question):
        message = question.topic

        buttons = []
        if question.options:
            for option in question.options:
                buttons.append(option)

        return {"message": message, "buttons": buttons}


class ViberService(ArchetypeService):
    def __init__(self, quiz, storage):
        super().__init__(quiz, storage)

    def run_forever(self):
        pass

    async def send_message(self, user_id, *args, **kwargs):
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['limpopo',
 'limpopo.services',
 'limpopo.storages',
 'limpopo.storages.postgres',
 'limpopo.storages.postgres.migrations',
 'limpopo.storages.postgres.migrations.versions']

package_data = \
{'': ['*']}

install_requires = \
['Telethon[pillow,cryptg]>=1.18.2,<2.0.0']

setup_kwargs = {
    'name': 'limpopo',
    'version': '0.1.2',
    'description': 'limpopo is a framework that allows you to create an application for conducting surveys via messengers',
    'long_description': 'Limpopo\n=======\n\n**limpopo** is a framework that allows you to create an application for conducting surveys in the following messengers:\n\n- Telegram\n\n- Viber\n\n\nInstalling\n----------\n\nInstall and update using `pip`:\n\n.. code-block:: text\n\n    pip install limpopo\n\n\nA Simple Example\n----------------\n\n.. code-block:: python\n\n    import env\n    import envparse  # External dependencies\n    import import\n    from limpopo.question import Question\n    from limpopo.services import TelegramService\n    from limpopo.storages import FakeStorage\n\n    how_are_you_question = Question(\n        topic="How are you?",\n        choices=[\n            "fine"\n        ],\n        strict_choose=False\n    )\n\n    await def quiz(dialog):\n        how_are_you_answer = await dialog.ask(how_are_you_question)\n\n        if how_are_you_answer.text != "fine":\n            await dialog.tell("Ohh")\n\n\n    def main():\n        settings = {\n            \'api_id\': env("TELEGRAM_API_ID", cast=int),\n            \'api_hash\': env("TELEGRAM_API_HASH", cast=str),\n            \'token\': env("TELEGRAM_BOT_TOKEN", cast=str)\n        }\n\n        storage = FakeStorage()\n\n        service = TelegramService(\n            quiz=quiz,\n            storage=storage,\n            settings=settings\n        )\n\n        service.run_forever()\n\nDesign\n------\n\nlimpopo provides the following entities, by which an poll-application is created:\n\n1. Service (limpopo provides `TelegramService`, `ViberService`)\n\n2. Storage (limpopo provides `PostgreStorage`, `FakeStorage`)\n\n3. Dialog\n\n4. Question\n\n\nDevelopment\n-----------\n\nHow to lint and format the code\n-------------------------------\n\nWe are using `pre-commit <https://pre-commit.com/>`_ tool,\nsee `installation guides <https://pre-commit.com/#installation>`_.\n\n.. code-block:: text\n\n    pre-commit install\n    pre-commit run -a\n',
    'author': 'limpopooooo',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0',
}


setup(**setup_kwargs)



ANSWER_TIMEOUT = 300  # 300 sec. timeout to answer in dialog

FOREWORD = "Для прохождения опроста, пожалуйста напишите /start"

CANCEL = "Опрос отменён! Если Вы передумаете в будущем, то начните опрос заново, используя команду /start."

WRONG_ANSWER_FORMAT = "Некорректный формат ответа!"

SESSION_DOESNT_EXIST = "Вы не проходили опрос, поэтому нечего останавливать"
